package dev.pfaff.altargui.io;

import java.io.IOException;
import java.lang.foreign.MemorySegment;
import java.nio.channels.ByteChannel;
import java.nio.channels.SeekableByteChannel;
import java.nio.file.Files;
import java.nio.file.Path;

public final class IOUtil {
	private static final int BUFFER_SIZE = 8192;

	public static MemorySegment readAllBytes(Reallocator allocator, Path path) throws IOException {
		try (SeekableByteChannel sbc = Files.newByteChannel(path)) {
			return read(allocator, sbc, sbc.size());
		}
	}

	private static MemorySegment read(Reallocator allocator, ByteChannel source, long initialSize) throws IOException {
		long capacity = initialSize;
		MemorySegment buf = allocator.reallocate(MemorySegment.NULL, capacity);
		long nread = 0;
		int n;
		// read to EOF which may read more or less than initialSize (eg: file
		// is truncated while we are reading)

		while ((n = source.read(buf.asSlice(nread, Math.min(buf.byteSize(), Integer.MAX_VALUE)).asByteBuffer())) != -1) {
			nread += n;
			if (nread == buf.byteSize()) {
				// need to allocate a larger buffer
				capacity = Math.max(newLength(capacity,
								1,       /* minimum growth */
								capacity /* preferred growth */),
						BUFFER_SIZE);
				buf = allocator.reallocate(buf, capacity);
			}
		}
		return buf.asSlice(nread);
	}

	private static long newLength(long oldLength, long minGrowth, long prefGrowth) {
		// preconditions not checked because of inlining
		// assert oldLength >= 0
		// assert minGrowth > 0

		long prefLength = oldLength + Math.max(minGrowth, prefGrowth); // might overflow
		if (prefLength >= 0) {
			return prefLength;
		} else {
			throw new OutOfMemoryError(STR."Required array length \{oldLength} + \{minGrowth} is too large");
		}
	}
}
